from . import applications, fems, meshes, solvers, tools

# from .__about__ import __author__, __author_email__, __version__, __website__
# __all__ = [
#     "__version__",
#     "__author__",
#     "__author_email__",
#     "__website__",
# ]
